maximumWeightSpace(0).
minimumWeightSpace(2).
maximumRockPushable(8).
majorForRockClimbing(4).
majorForAlreadyExisting(1000).
   
/* ------------------------------------------------
   Find a way  */
   
   % Basic path finding call
   % pathFinding(StartMap, Destination, Way)
   
   % Full path finding call
   % pathFinding(StartMap, Destination, Way, CanGoToNearTarget, MustBeAbleToLeave)
   
   % CanGoToNearTarget means we go to the first target we found (true or false)
   
   
% Basic path finding call
pathFinding(StartMap, Destination, Way)
:- pathFinding(StartMap, Destination, Way, true, true).
   
% Full path finding call
pathFinding(StartMap, Destination, Way, CanGoToNearTarget, MustBeAbleToLeave)
:- recorded(mapHasMoved, MapIsMoving),
   pathFinding(StartMap, MapIsMoving, Destination, Way, CanGoToNearTarget, MustBeAbleToLeave), !.
   
   
% Full path finding call
pathFinding(StartMap, MapIsMoving, Destination, Way, CanGoToNearTarget, MustBeAbleToLeave)
:- pathfinding_createCloseList(StartMap, MapIsMoving, Destination, CloseList),
   pathFinding([Destination, [], CloseList, [], 1, 1, CanGoToNearTarget, MustBeAbleToLeave, [], []], Info),
   pathFinding_getWay(Info, Way), !.



/* ------------------------------------------------
   Real path finding call */

% pathFinding(PathFindingInfos)

% PathFindingInfos : [Destination, OpenList, CloseList, WaitList, LastCaseId, LastMapId, CanGoToNearTarget, MustBeAbleToLeave, BreakPoint, Way]


% on arrive a la destination
pathFinding(Info, NewInfo)
:- pathFinding_getDestination(Info, Destination),
   pathFinding_getCloseList(Info, CloseList),
   pathFinding_getLastAddedMap(CloseList, Map),
   getCurrentCasePosInNavMap(Map, Destination),
   pathFinding_getLastAddedId(CloseList, DestinationId),
   pathFinding_computeWay(CloseList, DestinationId, Way),
   pathFinding_changeWay(Info, Way, NewInfo), !.
   
   
% boucle principale de pathfinding
pathFinding(Info, NewInfo)
:- debugN_pf('=== main PF loop ==='),
   pathFinding_addAdjacentCasesToOpenList(Info, Info2),
   debugN_pf('** adjacent added **'),
   pathFinding_dealBreakPoint_and_moveBestCaseFromOpenToCloseList(Info2, NewInfo), !.
   
   
/* ------------------------------------------------
   WHILE opration of main pathfinding loop */

   % pathFinding_dealBreakPoint_and_moveBestCaseFromOpenToCloseList(Info)

% BreakPoint and CanGoToNearTarget are true : we found an other target, just go there
pathFinding_dealBreakPoint_and_moveBestCaseFromOpenToCloseList(Info, NewInfo)
:- pathFinding_getCanGoToNearTarget(Info, true),
   pathFinding_getBreakPoint(Info, NewDestPos),
   isPosition(NewDestPos),
   
   debugN_pf('==== breakPoint ============================================'),
   
   pathFinding_getCloseList(Info, CloseList),
   pathFinding_getLastAddedId(CloseList, ParentId),
   
   debugN_pf('parent id'),
   
   pathFinding_getLastAddedMap(CloseList, Map),
   mapNavigateTo(Map, NewDestPos, NewMap),
   
   pathFinding_getLastCaseId(Info, LastId),
   NewId is LastId + 1,

   % ||||||||||||||||||||||||||||||||||||||||||||||||||||||
   % ||||||||||||||||||||||||||||||||||||||||||||||||||||||
   % ||||||||||||||||||||||||||||||||||||||||||||||||||||||
   % TODO MapIsMoving true or false
   

   pathFinding_addCaseInCloseList(CloseList, ParentId, NewId, NewMap, false, NewCloseList),
   
   pathFinding_computeWay(NewCloseList, NewId, Way),
   debug_pf('way found : '),
   debugN_pf(Way),
   
   pathFinding_changeWay(Info, Way, NewInfo),
   
   debug_pf('way : '),
   debugN_pf(Way), !.


% There is no breakpoint, add the best case from open to close list
pathFinding_dealBreakPoint_and_moveBestCaseFromOpenToCloseList(Info, NewInfo)
:- pathFinding_moveBestCaseFromOpenToCloseList(Info, Info2),
   pathFinding(Info2, NewInfo), !.
   


/* ------------------------------------------------
   A a case from open to close list */

   % pathFinding_moveBestCaseFromOpenToCloseList(OpenList, CloseList, NewOpenList, NewCloseList)

%pathFinding_moveBestCaseFromOpenToCloseList([ [ _ , Element ] | RestOpenList], CloseList, RestOpenList, [ Element | CloseList ] )
%:- debugN_pf('in pathFinding_moveBestCaseFromOpenToCloseList'), !.

pathFinding_moveBestCaseFromOpenToCloseList(Info, NewInfo)
:- pathFinding_getOpenList(Info, [ [ W , Element ] | RestOpenList]),
   pathFinding_getCloseList(Info, CloseList),
   
   pathFinding_changeOpenAndCloseLists(Info, RestOpenList, [ Element | CloseList ], NewInfo),
   
   % for tracing :
   pathFinding_getLastAddedMap([ Element | CloseList ], Map),
   getCurrentCasePosInNavMap(Map, Pos),
   debug_pf(' ///////////// case added : '),
   debug_pf(Pos),
   debug_pf('poid :'),
   debugN_pf(W), !.
   
   
pathFinding_moveBestCaseFromOpenToCloseList(Info, NewInfo)
:- pathFinding_getWaitList(Info, [ [ W , Element ] | RestWaitList]),
   debug_pf('Open List is empty, but there are'),
   getListSize(RestWaitList, NB),
   debug_pf(NB),
   debugN_pf('elements in Wait list'),

   pathFinding_getCloseList(Info, CloseList),

   pathFinding_changeWaitAndCloseLists(Info, RestWaitList, [ Element | CloseList ], NewInfo),

   % for tracing :
   pathFinding_getLastAddedMap([ Element | CloseList ], Map),
   getCurrentCasePosInNavMap(Map, Pos),
   debug_pf(' ///////////// case added : '),
   debug_pf(Pos),
   debug_pf('poid :'),
   debugN_pf(W), !.



/* ------------------------------------------------
   Rcuprer un chemin  partir d'une liste ferme  */

   % pathFinding_computeWay(CloseList, Destination, Way)

pathFinding_computeWay(CloseList, Destination, Way)
:- debug_pf('cherche comment arriver  :'),
   debugN_pf(Destination),
   %debugN_pf(CloseList),
   pathFinding_computeWay(CloseList, CloseList, Destination, [], Way), !.


% pathFinding_computeWay(CloseList, CloseList, Searched, CurrentWay, Way)

%[0, Dist, [-1, -1], NewNavMap]

% on est sur la case recherche et c'est la premiere case du chemin
pathFinding_computeWay(_, [ [_, _, 0, 1, _, _, Map | _] | _ ], 1, CurrentWay, [ Pos | CurrentWay ])
:- getCurrentCasePosInNavMap(Map, Pos), !.

% on est sur la case recherche
pathFinding_computeWay(CloseList, [ [_, _, Parent, Searched, _, _, Map | _] | _ ], Searched, CurrentWay, Way)
:- getCurrentCasePosInNavMap(Map, Pos),
   debug_pf('%% bonne case : '),
   debug_pf(Searched),
   debug_pf('parent :'),
   debugN_pf(Parent),
   pathFinding_computeWay(CloseList, CloseList, Parent, [ Pos | CurrentWay], Way), !.


% on n'est pas sur la case recherche
pathFinding_computeWay(CloseList, [ [_, _, _, Id | _] | Rest ], Searched, CurrentWay, Way)
:- debug_pf('mauvaise case : '),
   debugN_pf(Id),
   pathFinding_computeWay(CloseList, Rest, Searched, CurrentWay, Way), !.





/* ------------------------------------------------
   Calcul si la case ajoute devrait tre un breakpoint (diament ou cible permettant d'arreter la recherche de chemin) */

   % pathFinding_isBreakPoint(Map, CanGoToNearTarget, BreakPoint)


% la case est une cible
pathFinding_isBreakPoint(Map, true, BreakPoint)
:- getCurrentCaseInNavMap(Map, Case),
   caseIsTarget(Case),
   getCurrentCasePosInNavMap(Map, BreakPoint), !.


% la case n'est pas une cible ... pas de breakpoint
pathFinding_isBreakPoint(_, _, []) :- !.



/* ------------------------------------------------
   Rcuprer les coordonnes de la case parent */

   % pathFinding_getParentFromDirection(Pos, Direction, Parent)

pathFinding_getParentFromDirection([X, Y], up, [X, NewY])
:- NewY is Y + 1, !.

pathFinding_getParentFromDirection([X, Y], down, [X, NewY])
:- NewY is Y - 1, !.

pathFinding_getParentFromDirection([X, Y], left, [NewX, Y])
:- NewX is X + 1, !.

pathFinding_getParentFromDirection([X, Y], right, [NewX, Y])
:- NewX is X - 1, !.

pathFinding_getParentFromDirection([X, Y], here, [X, Y]) :- !.


/* ------------------------------------------------
   Calcul si on peut sortir de cette situation */

   % pathFinding_isAbleToLeave(Map, HasMovedMap, MustBeAbleToLeave)


% Si la carte n'a pas bouge, c'est bon
pathFinding_isAbleToLeave(_, false, _) :- !.


% si on a pas besoin de sortir, c'est bon
pathFinding_isAbleToLeave(_, _, false) :- !.


% sinon on cherche un chemin pr sortir
pathFinding_isAbleToLeave(Map, _, _)
:- getExitPos(Map, Destination),
   pathFinding(Map, Destination, _, false, false).


%.............. TODO ................
% creer le prdicat getExitPos(Map, Destination)



/* ------------------------------------------------
   Rcupre le premier break point disponible  */

   % pathFinding_dealBreakPoint(BreakPoint_1, BreakPoint_2, BreakPoint_3, BreakPoint_4, BreakPoint)

% Il n'y a pas de breakpoint
pathFinding_dealBreakPoint([], [], [], [], []) :- !.

% Il y a un breakpoint ... le renvoyer
pathFinding_dealBreakPoint([X,Y], _, _, _, [X,Y]) :- !.
pathFinding_dealBreakPoint(_, [X,Y], _, _, [X,Y]) :- !.
pathFinding_dealBreakPoint(_, _, [X,Y], _, [X,Y]) :- !.
pathFinding_dealBreakPoint(_, _, _, [X,Y], [X,Y]) :- !.


% Il n'y a pas de breakpoint
pathFinding_dealBreakPoint([], [], [], [], [], []) :- !.

% Il y a un breakpoint ... le renvoyer
pathFinding_dealBreakPoint([X,Y], _, _, _, _, [X,Y]) :- !.
pathFinding_dealBreakPoint(_, [X,Y], _, _, _, [X,Y]) :- !.
pathFinding_dealBreakPoint(_, _, [X,Y], _, _, [X,Y]) :- !.
pathFinding_dealBreakPoint(_, _, _, [X,Y], _, [X,Y]) :- !.
pathFinding_dealBreakPoint(_, _, _, _, [X,Y], [X,Y]) :- !.